//
//
//
#ifndef IO_H
#define IO_H

#define MAX_SCAN_VAL    11
#define MAX_EXP_GOLOMB  16

//
//  Bitstream
//
typedef struct
{
  int           frame_bitoffset;    //!< actual position in the codebuffer, bit-oriented, UVLC only
  int           bitstream_length;   //!< over codebuffer length, byte oriented, UVLC only
  byte          *streamBuffer;      //!< actual codebuffer for read bytes
} Bitstream;

//
// Syntaxelement
//
typedef struct syntaxelement
{
  int           type;                  //!< type of syntax element for data part.
  int           value1;                //!< numerical value of syntax element
  int           value2;                //!< for blocked symbols, e.g. run/level
  int           len;                   //!< length of code
  int           inf;                   //!< info part of UVLC code
  unsigned int  bitpattern;            //!< UVLC bitpattern  
  //! for mapping of UVLC to syntaxElement
  void    (*mapping)(int len, int info, int *value1, int *value2);
} SyntaxElement;


int read_one_new_frame(int height, int width, int skip, int start, int imgNo, FILE *p_in, imgStruct *img);
void write_image(int height, int width, FILE *p_out, imgStruct img);

Bitstream * allocBitstream(int size);
void freeBitstream(Bitstream *bitstream);
int readBitstream(Bitstream *bitstream, char *fname);

int u_1 (char *tracestring, Bitstream *bitstream);
int u_v (int LenInBits, char *tracestring, Bitstream *bitstream);
int readSyntaxElement_FLC(SyntaxElement *sym, Bitstream *currStream);
void linfo_ue(int len, int info, int *value1, int *dummy);
int golombDecode(int k, Bitstream *bitstream, char *tracestring);

void decodeAuxInfo(int *skipCoding, int *sqrFiltLength, int *filters_per_group, 
                   int varIndTab[NO_VAR_BINS], 
                   int *prevFiltInd, int usePrevFiltBest[NO_VAR_BINS], int frNo, Bitstream *bitstream);
void readPredFlags(int *force0, int *predMethod, int codedVarBins[NO_VAR_BINS], 
                   int filters_per_group, Bitstream *bitstream);
void readFilterCodingParams(int *minKStart, int kMinTab[], int sqrFiltLength, Bitstream *bitstream);
void readFilterCoeffs(int sqrFiltLength, int savedFilters, int **FilterCoeff, int kMinTab[], 
                      Bitstream *bitstream);
#endif // IO_H
